"""
Functions about miscellany.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely.geometry as sg

from . import constants as cs
from . import helpers as hp

# Help mypy but avoid circular imports
if TYPE_CHECKING:
    from .feed import Feed


def list_fields(feed: "Feed", table: str | None = None) -> pd.DataFrame:
    """
    Return a DataFrame describing all the fields of the GTFS tables in the given feed
    or in the given table if specified.

    The resulting DataFrame has the following columns.

    - ``'table'``: name of the GTFS table, e.g. ``'stops'``
    - ``'column'``: name of a column in the table,
      e.g. ``'stop_id'``
    - ``'num_values'``: number of values in the column
    - ``'num_nonnull_values'``: number of nonnull values in the
      column
    - ``'num_unique_values'``: number of unique values in the
      column, excluding null values
    - ``'min_value'``: minimum value in the column
    - ``'max_value'``: maximum value in the column

    If the table is not in the feed, then return an empty DataFrame
    If the table is not valid, raise a ValueError
    """
    gtfs_tables = list(cs.DTYPES.keys())
    if table is not None:
        if table not in gtfs_tables:
            raise ValueError(f"{table} is not a GTFS table")
        else:
            tables = [table]
    else:
        tables = gtfs_tables

    frames = []
    for table in tables:
        f = getattr(feed, table)
        if f is None:
            continue

        def my_agg(col):
            d = {}
            d["column"] = col.name
            d["num_values"] = col.size
            d["num_nonnull_values"] = col.count()
            d["num_unique_values"] = col.nunique()
            d["min_value"] = col.dropna().min()
            d["max_value"] = col.dropna().max()
            return pd.Series(d)

        g = f.apply(my_agg).T.reset_index(drop=True)
        g["table"] = table
        frames.append(g)

    cols = [
        "table",
        "column",
        "num_values",
        "num_nonnull_values",
        "num_unique_values",
        "min_value",
        "max_value",
    ]

    if not frames:
        f = pd.DataFrame()
    else:
        f = pd.concat(frames)
        # Rearrange columns
        f = f[cols].copy()

    return f


def describe(feed: "Feed", sample_date: str | None = None) -> pd.DataFrame:
    """
    Return a DataFrame of various feed indicators and values,
    e.g. number of routes.
    Specialize some those indicators to the given YYYYMMDD sample date string,
    e.g. number of routes active on the date.

    The resulting DataFrame has the columns

    - ``'indicator'``: string; name of an indicator, e.g. 'num_routes'
    - ``'value'``: value of the indicator, e.g. 27

    """
    from . import calendar as cl

    d = dict()
    dates = cl.get_dates(feed)
    d["agencies"] = feed.agency["agency_name"].tolist()
    d["timezone"] = feed.agency["agency_timezone"].iat[0]
    d["start_date"] = dates[0]
    d["end_date"] = dates[-1]
    d["num_routes"] = feed.routes.shape[0]
    d["num_trips"] = feed.trips.shape[0]
    d["num_stops"] = feed.stops.shape[0]
    if feed.shapes is not None:
        d["num_shapes"] = feed.shapes["shape_id"].nunique()
    else:
        d["num_shapes"] = 0

    if sample_date is None or sample_date not in feed.get_dates():
        sample_date = cl.get_first_week(feed)[3]
    d["sample_date"] = sample_date
    d["num_routes_active_on_sample_date"] = feed.get_routes(sample_date).shape[0]
    trips = feed.get_trips(sample_date)
    d["num_trips_active_on_sample_date"] = trips.shape[0]
    d["num_stops_active_on_sample_date"] = feed.get_stops(sample_date).shape[0]
    f = pd.DataFrame(list(d.items()), columns=["indicator", "value"])

    return f


def assess_quality(feed: "Feed") -> pd.DataFrame:
    """
    Return a DataFrame of various feed indicators and values,
    e.g. number of trips missing shapes.

    The resulting DataFrame has the columns

    - ``'indicator'``: string; name of an indicator, e.g. 'num_routes'
    - ``'value'``: value of the indicator, e.g. 27

    This function is odd but useful for seeing roughly how broken a feed is
    This function is not a GTFS validator.
    """
    d = dict()

    # Count duplicate route short names
    r = feed.routes
    dup = r.duplicated(subset=["route_short_name"])
    n = dup[dup].count()
    d["num_route_short_names_duplicated"] = n
    d["frac_route_short_names_duplicated"] = n / r.shape[0]

    # Count stop times missing shape_dist_traveled values
    st = feed.stop_times.sort_values(["trip_id", "stop_sequence"])
    if "shape_dist_traveled" in st.columns:
        # Count missing distances
        n = st[st["shape_dist_traveled"].isnull()].shape[0]
        d["num_stop_time_dists_missing"] = n
        d["frac_stop_time_dists_missing"] = n / st.shape[0]
    else:
        d["num_stop_time_dists_missing"] = st.shape[0]
        d["frac_stop_time_dists_missing"] = 1

    # Count direction_ids missing
    t = feed.trips
    if "direction_id" in t.columns:
        n = t[t["direction_id"].isnull()].shape[0]
        d["num_direction_ids_missing"] = n
        d["frac_direction_ids_missing"] = n / t.shape[0]
    else:
        d["num_direction_ids_missing"] = t.shape[0]
        d["frac_direction_ids_missing"] = 1

    # Count trips missing shapes
    if feed.shapes is not None:
        n = t[t["shape_id"].isnull()].shape[0]
    else:
        n = t.shape[0]
    d["num_trips_missing_shapes"] = n
    d["frac_trips_missing_shapes"] = n / t.shape[0]

    # Count missing departure times
    n = st[st["departure_time"].isnull()].shape[0]
    d["num_departure_times_missing"] = n
    d["frac_departure_times_missing"] = n / st.shape[0]

    # Count missing first departure times missing
    g = st.groupby("trip_id").first().reset_index()
    n = g[g["departure_time"].isnull()].shape[0]
    d["num_first_departure_times_missing"] = n
    d["frac_first_departure_times_missing"] = n / st.shape[0]

    # Count missing last departure times
    g = st.groupby("trip_id").last().reset_index()
    n = g[g["departure_time"].isnull()].shape[0]
    d["num_last_departure_times_missing"] = n
    d["frac_last_departure_times_missing"] = n / st.shape[0]

    # Opine
    if (
        (d["frac_first_departure_times_missing"] >= 0.1)
        or (d["frac_last_departure_times_missing"] >= 0.1)
        or d["frac_trips_missing_shapes"] >= 0.8
    ):
        d["assessment"] = "bad feed"
    elif (
        d["frac_direction_ids_missing"]
        or d["frac_stop_time_dists_missing"]
        or d["num_route_short_names_duplicated"]
    ):
        d["assessment"] = "probably a fixable feed"
    else:
        d["assessment"] = "good feed"

    f = pd.DataFrame(list(d.items()), columns=["indicator", "value"])

    return f


def convert_dist(feed: "Feed", new_dist_units: str) -> "Feed":
    """
    Convert the distances recorded in the ``shape_dist_traveled``
    columns of the given Feed to the given distance units.
    New distance units must lie in :const:`.constants.DIST_UNITS`.
    Return the resulting Feed.
    """
    feed = feed.copy()

    if feed.dist_units == new_dist_units:
        # Nothing to do
        return feed

    old_dist_units = feed.dist_units
    feed.dist_units = new_dist_units

    converter = hp.get_convert_dist(old_dist_units, new_dist_units)

    if hp.is_not_null(feed.stop_times, "shape_dist_traveled"):
        feed.stop_times["shape_dist_traveled"] = feed.stop_times[
            "shape_dist_traveled"
        ].map(converter)

    if hp.is_not_null(feed.shapes, "shape_dist_traveled"):
        feed.shapes["shape_dist_traveled"] = feed.shapes["shape_dist_traveled"].map(
            converter
        )

    return feed


def compute_network_stats_0(
    stop_times_subset: pd.DataFrame,
    trip_stats_subset: pd.DataFrame,
    *,
    split_route_types=False,
) -> pd.DataFrame:
    """
    Compute some network stats for the trips common to the
    given subset of stop times and given subses of trip stats
    of the form output by the function :func:`.trips.compute_trip_stats`

    Return a DataFrame with the columns

    - ``'route_type'`` (optional): presest if and only if ``split_route_types``
    - ``'num_stops'``: number of stops active on the date
    - ``'num_routes'``: number of routes active on the date
    - ``'num_trips'``: number of trips that start on the date
    - ``'num_trip_starts'``: number of trips with nonnull start
      times on the date
    - ``'num_trip_ends'``: number of trips with nonnull start times
      and nonnull end times on the date, ignoring trips that end
      after 23:59:59 on the date
    - ``'peak_num_trips'``: maximum number of simultaneous trips in
      service on the date
    - ``'peak_start_time'``: start time of first longest period
      during which the peak number of trips occurs on the date
    - ``'peak_end_time'``: end time of first longest period during
      which the peak number of trips occurs on the date
    - ``'service_distance'``: sum of the service distances for the
      active routes on the date;
      measured in kilometers if ``feed.dist_units`` is metric;
      otherwise measured in miles;
      contains all ``np.nan`` entries if ``feed.shapes is None``
    - ``'service_duration'``: sum of the service durations for the
      active routes on the date; measured in hours
    - ``'service_speed'``: service_distance/service_duration on the
      date

    Exclude dates with no active stops, which could yield the empty DataFrame.

    Helper function for :func:`compute_network_stats`.
    """
    final_cols = [
        "num_stops",
        "num_routes",
        "num_trips",
        "num_trip_starts",
        "num_trip_ends",
        "peak_num_trips",
        "peak_start_time",
        "peak_end_time",
        "service_distance",
        "service_duration",
        "service_speed",
    ]
    if split_route_types:
        final_cols.insert(0, "route_type")

    null_stats = pd.DataFrame(columns=final_cols)

    # Handle defunct case
    if stop_times_subset.empty or trip_stats_subset.empty:
        return null_stats

    # Handle generic case
    trip_ids = set(trip_stats_subset["trip_id"].values) & set(
        stop_times_subset["trip_id"].values
    )
    ts = trip_stats_subset.loc[lambda x: x["trip_id"].isin(trip_ids)].copy()
    st = stop_times_subset.loc[lambda x: x["trip_id"].isin(trip_ids)].copy()

    # Convert timestrings to seconds for quicker calculations
    ts[["start_time", "end_time"]] = ts[["start_time", "end_time"]].map(
        hp.timestr_to_seconds
    )

    def agg(g: pd.DataFrame, route_type: str | None = None) -> dict:
        d = {}
        if route_type is not None:
            d["route_type"] = route_type
        d["num_stops"] = st.loc[
            lambda x: x["trip_id"].isin(g["trip_id"]), "stop_id"
        ].nunique()
        d["num_routes"] = g["route_id"].nunique()
        d["num_trips"] = len(g)
        d["num_trip_starts"] = g["start_time"].count()
        d["num_trip_ends"] = g.loc[g["end_time"] < 24 * 3600, "end_time"].count()
        d["service_distance"] = g["distance"].sum()
        d["service_duration"] = g["duration"].sum()
        if d["service_distance"]:
            d["service_speed"] = d["service_distance"] / d["service_duration"]
        else:
            d["service_speed"] = 0

        # Compute peak stats, which is the slowest part
        active_trips = hp.get_active_trips_df(g[["start_time", "end_time"]])
        times, counts = (active_trips.index.values, active_trips.values)
        start, end = hp.get_peak_indices(times, counts)
        d["peak_num_trips"] = counts[start]
        d["peak_start_time"] = times[start]
        d["peak_end_time"] = times[end]
        return d

    # Compute stats
    if split_route_types:
        series = []
        for route_type, g in ts.groupby("route_type"):
            series.append(pd.Series(agg(g, route_type)))

        stats = pd.DataFrame(series)

    else:
        stats = pd.DataFrame(agg(ts), index=[0])

    # Convert seconds back to timestrings
    times = ["peak_start_time", "peak_end_time"]
    stats[times] = stats[times].map(lambda t: hp.seconds_to_timestr(t))

    return stats.filter(final_cols)


def compute_network_stats(
    feed: "Feed",
    dates: list[str],
    trip_stats: pd.DataFrame | None = None,
    *,
    split_route_types=False,
) -> pd.DataFrame:
    """
    Compute some network stats for the given subset of trip stats, which defaults to
    `feed.compute_trip_stats()`, and for the given dates (YYYYMMDD date stings).

    Return a table with the columns

    - ``'date'``
    - ``'route_type'`` (optional): presest if and only if ``split_route_types``
    - ``'num_stops'``: number of stops active on the date
    - ``'num_routes'``: number of routes active on the date
    - ``'num_trips'``: number of trips that start on the date
    - ``'num_trip_starts'``: number of trips with nonnull start
      times on the date
    - ``'num_trip_ends'``: number of trips with nonnull start times
      and nonnull end times on the date, ignoring trips that end
      after 23:59:59 on the date
    - ``'peak_num_trips'``: maximum number of simultaneous trips in
      service on the date
    - ``'peak_start_time'``: start time of first longest period
      during which the peak number of trips occurs on the date
    - ``'peak_end_time'``: end time of first longest period during
      which the peak number of trips occurs on the date
    - ``'service_distance'``: sum of the service distances for the
      active routes on the date;
      measured in kilometers if ``feed.dist_units`` is metric;
      otherwise measured in miles;
      contains all ``np.nan`` entries if ``feed.shapes is None``
    - ``'service_duration'``: sum of the service durations for the
      active routes on the date; measured in hours
    - ``'service_speed'``: service_distance/service_duration on the
      date

    Exclude dates with no active stops, which could yield the empty DataFrame.

    The route and trip stats for date d contain stats for trips that
    start on date d only and ignore trips that start on date d-1 and
    end on date d.

    Notes
    -----
    - If you've already computed trip stats in your workflow, then you should pass
      that table into this function to speed things up significantly.

    """
    dates = feed.subset_dates(dates)

    # Handle defunct case
    null_stats = compute_network_stats_0(
        pd.DataFrame(), pd.DataFrame(), split_route_types=split_route_types
    )
    if not dates:
        return null_stats

    final_cols = ["date"] + null_stats.columns.to_list()

    # Collect stats for each date,
    # memoizing stats the sequence of trip IDs active on the date
    # to avoid unnecessary recomputations.
    # Store in a dictionary of the form
    # trip ID sequence -> stats DataFrame.
    if trip_stats is None:
        trip_stats = feed.compute_trip_stats()

    activity = feed.compute_trip_activity(dates)
    stats_by_ids = {}
    frames = []
    for date in dates:
        ids = tuple(sorted(activity.loc[activity[date] > 0, "trip_id"].values))
        if ids in stats_by_ids:
            stats = (
                stats_by_ids[ids]
                # Assign date
                .assign(date=date)
            )
        elif ids:
            # Compute stats
            ts = trip_stats.loc[lambda x: x.trip_id.isin(ids)].copy()
            stats = (
                compute_network_stats_0(
                    feed.stop_times, ts, split_route_types=split_route_types
                )
                # Assign date
                .assign(date=date)
            )
            # Memoize stats
            stats_by_ids[ids] = stats
        else:
            stats = null_stats

        frames.append(stats)

    # Collate stats and order columns
    return pd.concat(frames, ignore_index=True).filter(final_cols)


def compute_network_time_series(
    feed: "Feed",
    dates: list[str],
    trip_stats: pd.DataFrame | None = None,
    freq: str = "h",
    *,
    split_route_types: bool = False,
) -> pd.DataFrame:
    """
    Compute some network stats in time series form for the given dates
    (YYYYMMDD date strings) and trip stats, which defaults to
    ``feed.compute_trip_stats()``.
    Use the given Pandas frequency string ``freq`` to specify the frequency of the
    resulting time series, e.g. '5Min'.
    If ``split_route_types``, then split stats by route type; otherwise don't.

    Return a long-form time series table with the columns

    - ``'datetime'``: datetime object
    - ``'route_type'``: integer; present if and only if ``split_route_types``
    - ``'num_trips'``: number of trips in service during during the
      time period
    - ``'num_trip_starts'``: number of trips with starting during the
      time period
    - ``'num_trip_ends'``: number of trips ending during the
      time period, ignoring the trips the end past midnight
    - ``'service_distance'``: distance traveled during the time
      period by all trips active during the time period;
      measured in kilometers if ``feed.dist_units`` is metric;
      otherwise measured in miles;
      contains all ``np.nan`` entries if ``feed.shapes is None``
    - ``'service_duration'``: duration traveled during the time
      period by all trips active during the time period;
      measured in hours
    - ``'service_speed'``: ``service_distance/service_duration`` when defined; 0
      otherwise

    Exclude dates that lie outside of the Feed's date range.
    If all the dates given lie outside of the Feed's date range,
    then return an empty DataFrame with the specified columns.

    Notes
    -----
    - If you've already computed trip stats in your workflow, then you should pass
      that table into this function to speed things up significantly.

    """
    final_cols = [
        "datetime",
        "num_trip_starts",
        "num_trip_ends",
        "num_trips",
        "service_distance",
        "service_duration",
        "service_speed",
    ]
    if split_route_types:
        final_cols.insert(1, "route_type")
    null_stats = pd.DataFrame(columns=final_cols)

    rts = feed.compute_route_time_series(dates, trip_stats, freq=freq)

    # Handle defunct case
    if rts.empty:
        return null_stats

    if trip_stats is None:
        trip_stats = feed.compute_trip_stats()

    if split_route_types:
        group_keys = ["datetime", "route_type"]
        rts = rts.merge(trip_stats.filter(["route_id", "route_type"]), how="left")
    else:
        group_keys = ["datetime"]

    indicators = set(final_cols) - {"datetime", "route_type"}
    f = (
        rts.groupby(group_keys)
        .agg(
            **{ind: (ind, lambda x: x.sum(min_count=1)) for ind in indicators}
        )  # All-NaNs should sum to NaN
        .reset_index()
    )
    # Recalculate service speed
    f["service_speed"] = (
        f["service_distance"]
        .div(f["service_duration"])
        .replace([np.inf, -np.inf], np.nan)
        .fillna(0.0)
    )

    return f


def create_shapes(feed: "Feed", *, all_trips: bool = False) -> "Feed":
    """
    Given a feed, create a shape for every trip that is missing a
    shape ID.
    Do this by connecting the stops on the trip with straight lines.
    Return the resulting feed which has updated shapes and trips
    tables.

    If ``all_trips``, then create new shapes for all trips by
    connecting stops, and remove the old shapes.
    """
    feed = feed.copy()

    if all_trips:
        trip_ids = feed.trips["trip_id"]
    else:
        trip_ids = feed.trips[feed.trips["shape_id"].isnull()]["trip_id"]

    # Get stop times for given trips
    f = feed.stop_times[feed.stop_times["trip_id"].isin(trip_ids)][
        ["trip_id", "stop_sequence", "stop_id"]
    ]
    f = f.sort_values(["trip_id", "stop_sequence"])

    if f.empty:
        # Nothing to do
        return feed

    # Create new shape IDs for given trips.
    # To do this, collect unique stop sequences,
    # sort them to impose a canonical order, and
    # assign shape IDs to them
    stop_seqs = sorted(
        set(tuple(group["stop_id"].values) for trip, group in f.groupby("trip_id"))
    )
    ids = hp.make_ids(len(stop_seqs), "shape_")
    shape_by_stop_seq = {seq: ids[i] for i, seq in enumerate(stop_seqs)}

    # Assign these new shape IDs to given trips
    shape_by_trip = {
        trip: shape_by_stop_seq[tuple(group["stop_id"].values)]
        for trip, group in f.groupby("trip_id")
    }
    trip_cond = feed.trips["trip_id"].isin(trip_ids)
    feed.trips.loc[trip_cond, "shape_id"] = feed.trips.loc[trip_cond, "trip_id"].map(
        lambda x: shape_by_trip[x]
    )

    # Build new shapes for given trips
    G = [
        [shape, i, stop]
        for stop_seq, shape in shape_by_stop_seq.items()
        for i, stop in enumerate(stop_seq)
    ]
    g = pd.DataFrame(G, columns=["shape_id", "shape_pt_sequence", "stop_id"])
    g = g.merge(feed.stops[["stop_id", "stop_lon", "stop_lat"]]).sort_values(
        ["shape_id", "shape_pt_sequence"]
    )
    g = g.drop(["stop_id"], axis=1)
    g = g.rename(columns={"stop_lon": "shape_pt_lon", "stop_lat": "shape_pt_lat"})

    if feed.shapes is not None and not all_trips:
        # Update feed shapes with new shapes
        feed.shapes = pd.concat([feed.shapes, g], sort=False)
    else:
        # Create all new shapes
        feed.shapes = g

    return feed


def compute_bounds(feed: "Feed", stop_ids: list[str] | None = None) -> np.array:
    """
    Return the bounding box (Numpy array [min longitude, min latitude, max longitude,
    max latitude]) of the given Feed's stops or of the subset of stops
    specified by the given stop IDs.
    """
    from .stops import get_stops

    g = get_stops(feed, as_gdf=True)
    if stop_ids is not None:
        g = g.loc[lambda x: x["stop_id"].isin(stop_ids)]

    return g.total_bounds


def compute_convex_hull(feed: "Feed", stop_ids: list[str] | None = None) -> sg.Polygon:
    """
    Return a convex hull (Shapely Polygon) representing the convex hull of the given
    Feed's stops or of the subset of stops specified by the given stop IDs.
    """
    from .stops import get_stops

    g = get_stops(feed, as_gdf=True)
    if stop_ids is not None:
        g = g.loc[lambda x: x["stop_id"].isin(stop_ids)]

    return g.union_all().convex_hull


def compute_centroid(feed: "Feed", stop_ids: list[str] | None = None) -> sg.Point:
    """
    Return the centroid (Shapely Point) of the convex hull the given Feed's stops
    or of the subset of stops specified by the given stop IDs.
    """
    from .stops import get_stops

    g = get_stops(feed, as_gdf=True)
    if stop_ids is not None:
        g = g.loc[lambda x: x["stop_id"].isin(stop_ids)]

    return g.union_all().convex_hull.centroid


def restrict_to_trips(feed: "Feed", trip_ids: list[str]) -> "Feed":
    """
    Build a new feed by restricting this one to only the stops,
    trips, shapes, etc. used by the trips of the given IDs.
    Return the resulting feed.

    If no valid trip IDs are given, which includes the case of the empty list,
    then the resulting feed will have all empty non-agency tables.

    This function is probably more useful internally than externally.
    """
    feed = feed.copy()
    has_agency_ids = "agency_id" in feed.routes.columns

    # Subset trips
    feed.trips = feed.trips.loc[lambda x: x.trip_id.isin(trip_ids)].copy()

    # Subset routes
    feed.routes = feed.routes.loc[lambda x: x.route_id.isin(feed.trips.route_id)].copy()

    # Subset stop times
    feed.stop_times = feed.stop_times.loc[lambda x: x.trip_id.isin(trip_ids)].copy()

    # Subset stops, collecting parent stations too
    stop_ids_0 = set(feed.stop_times["stop_id"])
    stop_ids_1 = set(
        feed.stops.loc[
            lambda x: x["stop_id"].isin(stop_ids_0), "parent_station"
        ].dropna()
    )
    stop_ids = stop_ids_0 | stop_ids_1
    feed.stops = feed.stops.loc[lambda x: x["stop_id"].isin(stop_ids)].copy()

    # Subset calendar
    service_ids = feed.trips["service_id"].unique()
    if feed.calendar is not None:
        feed.calendar = feed.calendar.loc[
            lambda x: x.service_id.isin(service_ids)
        ].copy()

    # Subset agency
    if has_agency_ids:
        agency_ids = feed.routes["agency_id"]
        feed.agency = feed.agency.loc[lambda x: x.agency_id.isin(agency_ids)].copy()

    # Now for the optional files.
    # Subset calendar dates.
    if feed.calendar_dates is not None:
        feed.calendar_dates = feed.calendar_dates.loc[
            lambda x: x.service_id.isin(service_ids)
        ].copy()

    # Subset frequencies
    if feed.frequencies is not None:
        feed.frequencies = feed.frequencies.loc[
            lambda x: x.trip_id.isin(trip_ids)
        ].copy()

    # Subset shapes
    if feed.shapes is not None:
        shape_ids = feed.trips.shape_id
        feed.shapes = feed.shapes.loc[lambda x: x.shape_id.isin(shape_ids)].copy()

    # Subset transfers
    if feed.transfers is not None:
        feed.transfers = feed.transfers.loc[
            lambda x: x.from_stop_id.isin(stop_ids) & x.to_stop_id.isin(stop_ids)
        ].copy()

    return feed


def restrict_to_routes(feed: "Feed", route_ids: list[str]) -> "Feed":
    """
    Build a new feed by restricting this one via :func:`restrict_to_trips` and
    the trips with the given route IDs.
    Return the resulting feed.
    """
    trip_ids = feed.trips.loc[lambda x: x.route_id.isin(route_ids), "trip_id"].tolist()
    return restrict_to_trips(feed, trip_ids)


def restrict_to_agencies(feed: "Feed", agency_ids: list[str]) -> "Feed":
    """
    Build a new feed by restricting this one via :func:`restrict_to_routes` and
    the routes with the given agency IDs.
    Return the resulting feed.
    """
    # Build feed via `restrict_to_routes`
    feed = feed.copy()
    route_ids = feed.routes.loc[
        lambda x: x.agency_id.isin(agency_ids), "route_id"
    ].tolist()

    return feed.restrict_to_routes(route_ids)


def restrict_to_dates(feed: "Feed", dates: list[str]) -> "Feed":
    """
    Build a new feed by restricting this one via :func:`restrict_to_trips` and
    the trips active on at least one of the given dates (YYYYMMDD strings).
    Return the resulting feed.
    """
    # Get every trip that is active on at least one of the dates
    trip_activity = feed.compute_trip_activity(dates)
    if trip_activity.empty:
        trip_ids = []
    else:
        trip_ids = trip_activity.loc[
            lambda x: x.filter(dates).sum(axis=1) > 0,
            "trip_id",
        ]

    return restrict_to_trips(feed, trip_ids)


def restrict_to_area(feed: "Feed", area: gpd.GeoDataFrame) -> "Feed":
    """
    Build a new feed by restricting this one via :func:`restrict_to_trips`
    and the trips that have at least one stop intersecting the given GeoDataFrame of
    polygons.
    Return the resulting feed.
    """
    from .stops import get_stops_in_area

    # Get IDs of stops within the polygon
    stop_ids = get_stops_in_area(feed, area).stop_id

    # Get all trips with at least one of those stops
    st = feed.stop_times.copy()
    trip_ids = st.loc[lambda x: x.stop_id.isin(stop_ids), "trip_id"]

    return restrict_to_trips(feed, trip_ids)


def _reshape_stop_times(stop_times: pd.DataFrame) -> pd.DataFrame:
    """
    Given a GTFS stop times DataFrame, reshape it to have only the following columns.

    - trip_id
    - stop_sequence
    - from_departure_time
    - to_departure_time
    - from_stop_id
    - to_stop_id
    - from_shape_dist_traveled (optional): present if and only if
      'shape_dist_traveled' column present in given stop times
    - to_shape_dist_traveled (optional): present if and only if 'shape_dist_traveled'
      column present in given stop times

    This is a helper function for :func:`compute_screen_line_counts`.
    """
    f = stop_times.sort_values(["trip_id", "stop_sequence"], ignore_index=True)
    g = f.groupby("trip_id")

    has_dist = "shape_dist_traveled" in f.columns

    # For each trip, create shifted columns for the "to" stop and its associated time and distance fields
    f["to_stop_id"] = g["stop_id"].shift(-1)
    f["to_departure_time"] = g["departure_time"].shift(-1)
    if has_dist:
        f["to_shape_dist_traveled"] = g["shape_dist_traveled"].shift(-1)

    # Drop rows where there is no "to" stop (i.e. the last stop in each trip)
    f = f.dropna(subset=["to_stop_id"])

    # Rename the original columns to reflect they represent the "from" stop in the segment
    f = f.rename(
        columns={
            "stop_id": "from_stop_id",
            "departure_time": "from_departure_time",
            "shape_dist_traveled": "from_shape_dist_traveled",
        }
    )

    return f.filter(
        [
            "trip_id",
            "stop_sequence",
            "from_departure_time",
            "to_departure_time",
            "from_stop_id",
            "to_stop_id",
            "from_shape_dist_traveled",
            "to_shape_dist_traveled",
        ]
    )


def compute_screen_line_counts(
    feed: "Feed",
    screen_lines: gpd.GeoDataFrame,
    dates: list[str],
    segmentize_m: float = 5,
    *,
    include_testing_cols: bool = False,
) -> pd.DataFrame:
    """
    Find all the Feed trips active on the given YYYYMMDD dates that intersect
    the given segment-associated screen lines of the form output by
    :func:`build_screen_lines`.
    Behind the scenes, use simple sub-LineStrings of the feed
    (with points separated by at ``segmentize_m`` meters)
    to compute screen line intersections.
    Using them instead of the Feed shapes avoids miscounting intersections in the
    case of non-simple (self-intersecting) shapes.

    For each trip crossing a screen line,
    compute the crossing time, crossing direction, etc. and return a DataFrame
    of results with the columns

    - ``'date'``: the YYYYMMDD date string given
    - ``'screen_line_id'``: ID of a screen line
    - ``'trip_id'``: ID of a trip that crosses the screen line
    - ``'shape_id'``: ID of the trip's shape
    - ``'direction_id'``: GTFS direction of trip
    - ``'route_id'``
    - ``'route_short_name'``
    - ``'route_type'``
    - ``'shape_id'``
    - ``'crossing_direction'``: 1 or -1; 1 indicates trip travel from the
      left side to the right side of the screen line;
      -1 indicates trip travel in the  opposite direction
    - ``'crossing_time'``: time, according to the GTFS schedule, that the trip
      crosses the screen line
    - ``'crossing_dist_m'``: distance along the trip shape (not subshape) of the
      crossing; in meters

    If ``include_testing_columns``, then include the following extra columns for testing
    purposes.

    - ``'subshape_id'``: ID of the simple sub-LineString S of the trip's shape that
      crosses the screen line
    - ``'subshape_length_m'``: length of S in meters
    - ``'from_departure_time'``: departure time of the trip from the last stop before
      the screen line
    - ``'to_departure_time'``: departure time of the trip at from the first stop after
      the screen line
    - ``'subshape_dist_frac'``: proportion of S's length at which the screen line
      intersects S

    Notes:

    - Assume the Feed's stop times DataFrame has an accurate ``shape_dist_traveled``
      column.
    - Assume that trips travel in the same direction as their shapes, an assumption
      that is part of the GTFS.
    - Assume that the screen line is straight and simple.
    - The algorithm works as follows

        1. Find the Feed's simple subshapes (computed via :func:`shapes.split_simple`)
           that intersect the screen lines.
        2. For each such subshape and screen line, compute the intersection points,
           the distance of each point along the subshape, aka the *crossing distance*,
           and the orientation of the screen line relative to the subshape.
        3. Restrict to trips active on the given dates and for each trip associated to
           an intersecting subshape above, interpolate a trip stop time
           for the intersection point using the crossing distance, subshape length,
           cumulative subshape length, and trip stop times.

    """
    from .shapes import split_simple

    # Convert geoms to UTM
    crs = screen_lines.estimate_utm_crs()
    screen_lines = screen_lines.to_crs(crs)

    # Create screen line IDs if necessary
    n = screen_lines.shape[0]
    if "screen_line_id" not in screen_lines.columns:
        screen_lines["screen_line_id"] = hp.make_ids(n, "sl")

    # Make a vector in the direction of each screen line to calculate crossing
    # orientation. Does not work in case of a bent screen line.
    p1 = screen_lines["geometry"].map(lambda x: np.array(x.coords[0]))
    p2 = screen_lines["geometry"].map(lambda x: np.array(x.coords[-1]))
    screen_lines["screen_line_vector"] = p2 - p1

    # Get the simple subshapes that intersect the screen lines.
    # Need subshapes to have only small gaps between them,
    # so `segmentize_m` needs to be small.
    subshapes = (
        feed.get_shapes(as_gdf=True, use_utm=True)
        .sjoin(screen_lines)
        .drop_duplicates("shape_id")
        .pipe(split_simple, segmentize_m=segmentize_m)
    )

    # Get intersection points of subshapes and screen lines
    g0 = (
        subshapes.sjoin(screen_lines.filter(["screen_line_id", "geometry"]))
        .merge(screen_lines, on="screen_line_id")
        .assign(
            int_point=lambda x: gpd.GeoSeries(x["geometry_x"], crs=crs).intersection(
                gpd.GeoSeries(x["geometry_y"], crs=crs)
            )
        )
    )

    # Unpack multipoint intersections to yield a new GeoDataFrame.
    # Should be very few multipoints.
    records = []
    for row in g0.itertuples(index=False):
        if isinstance(row.int_point, sg.Point):
            intersections = [row.int_point]
        else:
            intersections = row.int_point.geoms
        for int_point in intersections:
            record = {
                "subshape_id": row.subshape_id,
                "shape_id": row.shape_id,
                "subshape_length_m": row.subshape_length_m,
                "cum_length_m": row.cum_length_m,
                "screen_line_id": row.screen_line_id,
                "geometry": row.geometry_x,
                "int_point": int_point,
                "screen_line_vector": row.screen_line_vector,
            }
            records.append(record)

    g = gpd.GeoDataFrame.from_records(records).set_geometry("geometry").set_crs(crs)

    # Get distance (in meters) of each intersection point along subshape
    g["subshape_dist_frac"] = g.apply(
        lambda x: x["geometry"].project(x.int_point, normalized=True), axis=1
    )
    g["subshape_dist_m"] = g["subshape_dist_frac"] * g["subshape_length_m"]
    g["crossing_dist_m"] = (
        g["subshape_dist_m"] + g["cum_length_m"] - g["subshape_length_m"]
    )

    # Build a tiny vector along each subshape from the intersection point
    p2 = g.apply(
        lambda x: x["geometry"].interpolate(x["subshape_dist_m"] + 1), axis=1
    ).map(lambda x: np.array(x.coords[0]))
    p1 = g.int_point.map(lambda x: np.array(x.coords[0]))
    g["subshape_vector"] = p2 - p1

    # Compute crossing direction by taking the vector cross product of
    # the link vector and the screen line vector
    det = g.apply(
        lambda x: np.linalg.det(
            np.array([x["subshape_vector"], x["screen_line_vector"]])
        ),
        axis=1,
    )
    g["crossing_direction"] = det.map(lambda x: 1 if x >= 0 else -1)

    # Summarize work so far
    g = g[
        [
            "subshape_id",
            "shape_id",
            "screen_line_id",
            "subshape_dist_frac",
            "subshape_dist_m",
            "subshape_length_m",
            "crossing_direction",
            "crossing_dist_m",
        ]
    ]

    # Get stop times to compute crossing times
    feed = feed.convert_dist("m")
    frames = []
    for date in dates:
        st = (
            feed.get_stop_times(date)
            .pipe(_reshape_stop_times)
            .merge(feed.trips[["trip_id", "shape_id"]])
            # Keep only non-NaN departure times
            .loc[lambda x: x["from_departure_time"].notna()]
            .loc[lambda x: x["to_departure_time"].notna()]
            # Convert to seconds past midnight for upcoming crossing time calculation
            .assign(
                t1=lambda x: x["from_departure_time"].map(hp.timestr_to_seconds),
                t2=lambda x: x["to_departure_time"].map(hp.timestr_to_seconds),
            )
        )

        # Compute crossing times
        subframes = []
        for shape_id, group in g.groupby("shape_id"):
            f = (
                st.merge(group)
                # Only keep the times of the pair of stops on either side of each screen line,
                # whose distance along a trip shape is marked by column 'crossing_dist_m'
                .loc[lambda x: x["from_shape_dist_traveled"] <= x["crossing_dist_m"]]
                .loc[lambda x: x["crossing_dist_m"] <= x["to_shape_dist_traveled"]]
            )
            f["crossing_time"] = (
                f["t1"] + f["subshape_dist_frac"] * (f["t2"] - f["t1"])
            ).map(lambda x: hp.seconds_to_timestr(x))
            # Get distance along trip shape of crossing point
            subframes.append(f)

        if subframes:
            f = pd.concat(subframes).assign(date=date)
        else:
            f = pd.DataFrame()
        frames.append(f)

    f = pd.concat(frames)

    # Clean up
    final_cols = [
        "date",
        "segment_id",
        "segment_length",
        "screen_line_id",
        "shape_id",
        "trip_id",
        "direction_id",
        "route_id",
        "route_short_name",
        "route_type",
        "crossing_direction",
        "crossing_time",
        "crossing_dist_m",
    ]
    if include_testing_cols:
        final_cols += [
            "subshape_id",
            "subshape_length_m",
            "from_departure_time",
            "to_departure_time",
            "subshape_dist_frac",
            "subshape_dist_m",
        ]

    return (
        f
        # Append screen line info
        .merge(screen_lines.drop("geometry", axis=1))
        # Append extra trip info
        .merge(feed.trips[["trip_id", "direction_id", "route_id"]])
        .merge(feed.routes[["route_id", "route_short_name", "route_type"]])
        .filter(final_cols)
        .sort_values(
            ["screen_line_id", "trip_id", "crossing_dist_m"],
            ignore_index=True,
        )
    )



"""
Functions about trips.
"""

from __future__ import annotations

import datetime as dt
import functools as ft
import json
from typing import TYPE_CHECKING, Iterable

import folium as fl
import folium.plugins as fp
import numpy as np
import pandas as pd
import shapely.geometry as sg
import shapely.ops as so

from . import constants as cs
from . import helpers as hp

# Help mypy but avoid circular imports
if TYPE_CHECKING:
    from .feed import Feed


def get_active_services(feed: "Feed", date: str) -> list[str]:
    """
    Given a Feed and a date string in YYYYMMDD format,
    return the list of service IDs that are active on the date.
    """
    active_services = set()
    removed_services = set()

    if feed.calendar is not None:
        # Convert date to weekday string (e.g., "monday")
        weekday_str = dt.datetime.strptime(date, "%Y%m%d").strftime("%A").lower()
        # Filter `calendar` to services active on date
        active_services |= set(
            feed.calendar.loc[
                lambda x: (x["start_date"] <= date)
                & (x["end_date"] >= date)
                & (x[weekday_str] == 1),
                "service_id",
            ]
        )

    if feed.calendar_dates is not None:
        # Filter `calendar_dates` to services added on date
        active_services |= set(
            feed.calendar_dates.loc[
                lambda x: (x["date"] == date) & (x["exception_type"] == 1),
                "service_id",
            ]
        )
        # Filter `calendar_dates` to services removed on date
        removed_services |= set(
            feed.calendar_dates.loc[
                lambda x: (x["date"] == date) & (x["exception_type"] == 2),
                "service_id",
            ]
        )

    return list(active_services - removed_services)


def get_trips(
    feed: "Feed",
    date: str | None = None,
    time: str | None = None,
    *,
    as_gdf: bool = False,
    use_utm: bool = False,
) -> pd.DataFrame:
    """
    Return ``feed.trips``.
    If date (YYYYMMDD date string) is given then subset the result to trips
    that start on that date.
    If a time (HH:MM:SS string, possibly with HH > 23) is given in addition to a date,
    then further subset the result to trips in service at that time.

    If ``as_gdf`` and ``feed.shapes`` is not None, then return the trips as a
    GeoDataFrame of LineStrings representating trip shapes.
    Use local UTM CRS if ``use_utm``; otherwise it the WGS84 CRS.
    If ``as_gdf`` and ``feed.shapes`` is ``None``, then raise a ValueError.
    """
    if feed.trips is None:
        return None

    f = feed.trips
    if date is not None:
        f = f.loc[lambda x: x["service_id"].isin(get_active_services(feed, date))]

        if time is not None:
            # Get trips active during given time
            def get_active(group):
                d = {}
                start = group["departure_time"].dropna().min()
                end = group["departure_time"].dropna().max()
                try:
                    result = start <= time <= end
                except TypeError:
                    result = False
                d["is_active"] = result
                return pd.Series(d)

            f = (
                f.merge(feed.stop_times[["trip_id", "departure_time"]])
                .groupby("trip_id")
                .apply(get_active, include_groups=False)
                .reset_index()
                .loc[lambda x: x["is_active"]]
                .merge(f)
                .drop("is_active", axis=1)
            )

    if as_gdf:
        if feed.shapes is None:
            raise ValueError("This Feed has no shapes.")
        else:
            from .shapes import get_shapes

            f = (
                get_shapes(feed, as_gdf=True, use_utm=use_utm)
                .filter(["shape_id", "geometry"])
                .merge(f, how="right")
                .filter(f.columns.tolist() + ["geometry"])
            )

    return f


def compute_trip_activity(feed: "Feed", dates: list[str]) -> pd.DataFrame:
    """
    Mark trips as active or inactive on the given dates (YYYYMMDD date strings).
    Return a table with the columns

    - ``'trip_id'``
    - ``dates[0]``: 1 if the trip is active on ``dates[0]``;
      0 otherwise
    - ``dates[1]``: 1 if the trip is active on ``dates[1]``;
      0 otherwise
    - etc.
    - ``dates[-1]``: 1 if the trip is active on ``dates[-1]``;
      0 otherwise

    If ``dates`` is ``None`` or the empty list, then return an
    empty DataFrame.
    """
    dates = feed.subset_dates(dates)
    if not dates:
        return pd.DataFrame()

    # Get trip activity table for each day
    frames = [feed.trips[["trip_id"]]]
    for date in dates:
        frames.append(get_trips(feed, date)[["trip_id"]].assign(**{date: 1}))

    # Merge daily trip activity tables into a single table
    f = ft.reduce(lambda left, right: left.merge(right, how="outer"), frames).fillna(
        {date: 0 for date in dates}
    )
    f[dates] = f[dates].astype(int)
    return f


def compute_busiest_date(feed: "Feed", dates: list[str]) -> str:
    """
    Given a list of dates (YYYYMMDD date strings), return the first date that has the
    maximum number of active trips.
    """
    f = feed.compute_trip_activity(dates)
    s = [(f[c].sum(), c) for c in f.columns if c != "trip_id"]
    return max(s)[1]


def name_stop_patterns(feed: "Feed") -> pd.DataFrame:
    """
    For each (route ID, direction ID) pair, find the distinct stop patterns of its
    trips, and assign them each an integer *pattern rank* based on the stop pattern's
    frequency rank, where 1 is the most frequent stop pattern, 2 is the second most
    frequent, etc.
    Return the DataFrame ``feed.trips`` with the additional column
    ``stop_pattern_name``, which equals the trip's 'direction_id' concatenated with a
    dash and its stop pattern rank.

    If ``feed.trips`` has no 'direction_id' column, then temporarily create one equal
    to all zeros, proceed as above, then delete the column.
    """
    # Collect stop patterns
    trips = feed.trips.copy()
    if "direction_id" in trips.columns:
        has_direction = True
    else:
        has_direction = False
        trips["direction_id"] = 0  # placeholder to ease calculations below
    f = (
        feed.stop_times.sort_values(["trip_id", "stop_sequence"])
        .groupby("trip_id")
        .apply(
            lambda x: pd.Series({"stop_pattern": "-".join(x["stop_id"])}),
            include_groups=False,
        )
        .reset_index()
        .merge(trips[["route_id", "trip_id", "direction_id"]])
    )
    # Compute pattern ranks
    f["rank"] = f.groupby(["route_id", "direction_id"])["stop_pattern"].transform(
        lambda x: pd.Series(x).map(
            {pattern: idx + 1 for idx, pattern in enumerate(x.value_counts().index)}
        )
    )
    # Assign pattern names
    f["stop_pattern_name"] = (
        f["direction_id"].astype(str).str.cat(f["rank"].astype(str), sep="-")
    )
    ff = trips.merge(f[["route_id", "trip_id", "direction_id", "stop_pattern_name"]])
    # Clean up
    if not has_direction:
        del ff["direction_id"]

    return ff


def compute_trip_stats(
    feed: "Feed",
    route_ids: list[str | None] = None,
    *,
    compute_dist_from_shapes: bool = False,
) -> pd.DataFrame:
    """
    Return a DataFrame with the following columns:

    - ``'trip_id'``
    - ``'route_id'``
    - ``'route_short_name'``
    - ``'route_type'``
    - ``'direction_id'``: NaN if missing from feed
    - ``'shape_id'``: NaN if missing from feed
    - ``'stop_pattern_name'``: output from :func:`name_stop_patterns`
    - ``'num_stops'``: number of stops on trip
    - ``'start_time'``: first departure time of the trip
    - ``'end_time'``: last departure time of the trip
    - ``'start_stop_id'``: stop ID of the first stop of the trip
    - ``'end_stop_id'``: stop ID of the last stop of the trip
    - ``'is_loop'``: 1 if the start and end stop are less than 400m apart and
      0 otherwise
    - ``'distance'``: distance of the trip;
      measured in kilometers if ``feed.dist_units`` is metric;
      otherwise measured in miles;
      contains all ``np.nan`` entries if ``feed.shapes is None``
    - ``'duration'``: duration of the trip in hours
    - ``'speed'``: distance/duration

    If ``feed.stop_times`` has a ``shape_dist_traveled`` column with at
    least one non-NaN value and ``compute_dist_from_shapes == False``,
    then use that column to compute the distance column.
    Else if ``feed.shapes is not None``, then compute the distance
    column using the shapes and Shapely.
    Otherwise, set the distances to NaN.

    If route IDs are given, then restrict to trips on those routes.

    Notes
    -----
    - Assume the following feed attributes are not ``None``:

        * ``feed.trips``
        * ``feed.routes``
        * ``feed.stop_times``
        * ``feed.shapes`` (optionally)

    - Calculating trip distances with ``compute_dist_from_shapes=True``
      seems pretty accurate.  For example, calculating trip distances on
      `this Portland feed
      <https://transitfeeds.com/p/trimet/43/1400947517>`_
      using ``compute_dist_from_shapes=False`` and
      ``compute_dist_from_shapes=True``,
      yields a difference of at most 0.83km from the original values.

    """
    f = name_stop_patterns(feed)

    # Restrict to given route IDs
    if route_ids is not None:
        f = f[f["route_id"].isin(route_ids)].copy()

    # Merge with stop times and extra trip info.
    # Convert departure times to seconds past midnight to
    # compute trip durations later.
    if "direction_id" not in f.columns:
        f["direction_id"] = np.nan
    if "shape_id" not in f.columns:
        f["shape_id"] = np.nan

    f = (
        f[["route_id", "trip_id", "direction_id", "shape_id", "stop_pattern_name"]]
        .merge(feed.routes[["route_id", "route_short_name", "route_type"]])
        .merge(feed.stop_times)
        .sort_values(["trip_id", "stop_sequence"])
        .assign(departure_time=lambda x: x["departure_time"].map(hp.timestr_to_seconds))
    )

    # Compute all trips stats except distance,
    # which is possibly more involved
    geometry_by_stop = feed.build_geometry_by_stop(use_utm=True)
    g = f.groupby("trip_id")

    def my_agg(group):
        d = dict()
        d["route_id"] = group["route_id"].iat[0]
        d["route_short_name"] = group["route_short_name"].iat[0]
        d["route_type"] = group["route_type"].iat[0]
        d["direction_id"] = group["direction_id"].iat[0]
        d["shape_id"] = group["shape_id"].iat[0]
        d["stop_pattern_name"] = group["stop_pattern_name"].iat[0]
        d["num_stops"] = group.shape[0]
        d["start_time"] = group["departure_time"].iat[0]
        d["end_time"] = group["departure_time"].iat[-1]
        d["start_stop_id"] = group["stop_id"].iat[0]
        d["end_stop_id"] = group["stop_id"].iat[-1]
        dist = geometry_by_stop[d["start_stop_id"]].distance(
            geometry_by_stop[d["end_stop_id"]]
        )
        d["is_loop"] = int(dist < 400)
        d["duration"] = (d["end_time"] - d["start_time"]) / 3600
        return pd.Series(d)

    # Apply my_agg, but don't reset index yet.
    # Need trip ID as index to line up the results of the
    # forthcoming distance calculation
    h = g.apply(my_agg, include_groups=False)

    # Compute distance
    if hp.is_not_null(f, "shape_dist_traveled") and not compute_dist_from_shapes:
        # Compute distances using shape_dist_traveled column, converting to km or mi
        if hp.is_metric(feed.dist_units):
            convert_dist = hp.get_convert_dist(feed.dist_units, "km")
        else:
            convert_dist = hp.get_convert_dist(feed.dist_units, "mi")
        h["distance"] = g.apply(
            lambda group: convert_dist(group.shape_dist_traveled.max())
        )
    elif feed.shapes is not None:
        # Compute distances using the shapes and Shapely
        geometry_by_shape = feed.build_geometry_by_shape(use_utm=True)
        # Convert to km or mi
        if hp.is_metric(feed.dist_units):
            m_to_dist = hp.get_convert_dist("m", "km")
        else:
            m_to_dist = hp.get_convert_dist("m", "mi")

        def compute_dist(group):
            """
            Return the distance traveled along the trip between the
            first and last stops.
            If that distance is negative or if the trip's linestring
            intersects itfeed, then return the length of the trip's
            linestring instead.
            """
            shape = group["shape_id"].iat[0]
            try:
                # Get the linestring for this trip
                linestring = geometry_by_shape[shape]
            except KeyError:
                # Shape ID is NaN or doesn't exist in shapes.
                # No can do.
                return np.nan

            # If the linestring intersects itfeed, then that can cause
            # errors in the computation below, so just
            # return the length of the linestring as a good approximation
            D = linestring.length
            if not linestring.is_simple:
                return D

            # Otherwise, return the difference of the distances along
            # the linestring of the first and last stop
            start_stop = group["stop_id"].iat[0]
            end_stop = group["stop_id"].iat[-1]
            try:
                start_point = geometry_by_stop[start_stop]
                end_point = geometry_by_stop[end_stop]
            except KeyError:
                # One of the two stop IDs is NaN, so just
                # return the length of the linestring
                return D
            d1 = linestring.project(start_point)
            d2 = linestring.project(end_point)
            d = d2 - d1
            if 0 < d < D + 100:
                return d
            else:
                # Something is probably wrong, so just
                # return the length of the linestring
                return D

        h["distance"] = g.apply(compute_dist, include_groups=False)
        # Convert from meters
        h["distance"] = h["distance"].map(m_to_dist)
    else:
        h["distance"] = np.nan

    # Reset index and compute final stats
    h = h.reset_index()
    h["speed"] = h["distance"] / h["duration"]
    h[["start_time", "end_time"]] = h[["start_time", "end_time"]].map(
        lambda x: hp.seconds_to_timestr(x)
    )

    return h.sort_values(["route_id", "direction_id", "start_time"], ignore_index=True)


def locate_trips(feed: "Feed", date: str, times: list[str]) -> pd.DataFrame:
    """
    Return the positions of all trips active on the
    given date (YYYYMMDD date string) and times (HH:MM:SS time strings,
    possibly with HH > 23).

    Return a DataFrame with the columns

    - ``'trip_id'``
    - ``'route_id'``
    - ``'direction_id'``: all NaNs if ``feed.trips.direction_id`` is
      missing
    - ``'time'``
    - ``'rel_dist'``: number between 0 (start) and 1 (end)
      indicating the relative distance of the trip along its path
    - ``'lon'``: longitude of trip at given time
    - ``'lat'``: latitude of trip at given time

    Assume ``feed.stop_times`` has an accurate
    ``shape_dist_traveled`` column.
    """
    if not hp.is_not_null(feed.stop_times, "shape_dist_traveled"):
        raise ValueError(
            "feed.stop_times needs to have a non-null shape_dist_traveled "
            "column. You can create it, possibly with some inaccuracies, "
            "via feed2 = feed.append_dist_to_stop_times()."
        )

    if "shape_id" not in feed.trips.columns:
        raise ValueError("feed.trips.shape_id must exist.")

    # Start with stop times active on date
    f = feed.get_stop_times(date)
    f["departure_time"] = f["departure_time"].map(hp.timestr_to_seconds)

    # Compute relative distance of each trip along its path
    # at the given time times.
    # Use linear interpolation based on stop departure times and
    # shape distance traveled.
    geometry_by_shape = feed.build_geometry_by_shape(use_utm=False)
    sample_times = np.array([hp.timestr_to_seconds(s) for s in times])

    def compute_rel_dist(group):
        dists = sorted(group["shape_dist_traveled"].values)
        times = sorted(group["departure_time"].values)
        ts = sample_times[(sample_times >= times[0]) & (sample_times <= times[-1])]
        ds = np.interp(ts, times, dists)
        return pd.DataFrame({"time": ts, "rel_dist": ds / dists[-1]})

    # return f.groupby('trip_id', group_keys=False).\
    #   apply(compute_rel_dist).reset_index()
    g = f.groupby("trip_id").apply(compute_rel_dist, include_groups=False).reset_index()

    # Delete extraneous multi-index column
    del g["level_1"]

    # Convert times back to time strings
    g["time"] = g["time"].map(lambda x: hp.seconds_to_timestr(x))

    # Merge in more trip info and
    # compute longitude and latitude of trip from relative distance
    t = feed.trips.copy()
    if "direction_id" not in t.columns:
        t["direction_id"] = np.nan

    h = pd.merge(g, t[["trip_id", "route_id", "direction_id", "shape_id"]])
    if not h.shape[0]:
        # Return a DataFrame with the promised headers but no data.
        # Without this check, result below could be an empty DataFrame.
        h["lon"] = pd.Series()
        h["lat"] = pd.Series()
        return h

    def get_lonlat(group):
        shape = group.name
        linestring = geometry_by_shape[shape]
        lonlats = [
            linestring.interpolate(d, normalized=True).coords[0]
            for d in group["rel_dist"].values
        ]
        group["lon"], group["lat"] = zip(*lonlats)
        return group

    return (
        h.groupby("shape_id")
        .apply(get_lonlat, include_groups=False)
        .reset_index()
        .drop("level_1", axis=1)  # Where did this column come from?
    )


def trips_to_geojson(
    feed: "Feed",
    trip_ids: Iterable[str] | None = None,
    *,
    include_stops: bool = False,
) -> dict:
    """
    Return a GeoJSON FeatureCollection of LineString features representing
    all the Feed's trips.
    The coordinates reference system is the default one for GeoJSON,
    namely WGS84.

    If ``include_stops``, then include the trip stops as Point features.
    If an iterable of trip IDs is given, then subset to those trips.
    If any of the given trip IDs are not found in the feed, then raise a ValueError.
    If the Feed has no shapes, then raise a ValueError.
    """
    if trip_ids is None or not list(trip_ids):
        trip_ids = feed.trips.trip_id

    D = set(trip_ids) - set(feed.trips.trip_id)
    if D:
        raise ValueError(f"Trip IDs {D} not found in feed.")

    # Get trips
    g = get_trips(feed, as_gdf=True).loc[lambda x: x["trip_id"].isin(trip_ids)]
    trips_gj = json.loads(g.to_json())

    # Get stops if desired
    if include_stops:
        st_gj = feed.stop_times_to_geojson(trip_ids)
        trips_gj["features"].extend(st_gj["features"])

    return hp.drop_feature_ids(trips_gj)


def map_trips(
    feed: "Feed",
    trip_ids: Iterable[str],
    color_palette: list[str] = cs.COLORS_SET2,
    *,
    show_stops: bool = False,
    show_direction: bool = False,
):
    """
    Return a Folium map showing the given trips and (optionally)
    their stops.
    If any of the given trip IDs are not found in the feed, then raise a ValueError.
    If ``include_direction``, then use the Folium plugin PolyLineTextPath to draw arrows
    on each trip polyline indicating its direction of travel; this fails to work in some
    browsers, such as Brave 0.68.132.
    """
    # Initialize map
    my_map = fl.Map(tiles="cartodbpositron")

    # Create colors
    n = len(trip_ids)
    colors = [color_palette[i % len(color_palette)] for i in range(n)]

    # Collect bounding boxes to set map zoom later
    bboxes = []

    # Create a feature group for each route and add it to the map
    for i, trip_id in enumerate(trip_ids):
        collection = trips_to_geojson(feed, [trip_id], include_stops=show_stops)

        group = fl.FeatureGroup(name=f"Trip {trip_id}")
        color = colors[i]

        for f in collection["features"]:
            prop = f["properties"]

            # Add stop if present
            if f["geometry"]["type"] == "Point":
                lon, lat = f["geometry"]["coordinates"]
                fl.CircleMarker(
                    location=[lat, lon],
                    radius=8,
                    fill=True,
                    color=color,
                    weight=1,
                    popup=fl.Popup(hp.make_html(prop)),
                ).add_to(group)

            # Add trip
            else:
                path = fl.PolyLine(
                    [[x[1], x[0]] for x in f["geometry"]["coordinates"]],
                    color=color,
                    popup=hp.make_html(prop),
                )

                path.add_to(group)
                bboxes.append(sg.box(*sg.shape(f["geometry"]).bounds))

                if show_direction:
                    # Direction arrows, assuming, as GTFS does, that
                    # trip direction equals LineString direction
                    fp.PolyLineTextPath(
                        path,
                        "        \u27a4        ",
                        repeat=True,
                        offset=5.5,
                        attributes={"fill": color, "font-size": "18"},
                    ).add_to(group)

        group.add_to(my_map)

    fl.LayerControl().add_to(my_map)

    # Fit map to bounds
    bounds = so.unary_union(bboxes).bounds
    # Folium wants a different ordering
    bounds = [(bounds[1], bounds[0]), (bounds[3], bounds[2])]
    my_map.fit_bounds(bounds)

    return my_map



"""
Functions about shapes.
"""

from __future__ import annotations
from typing import Iterable, TYPE_CHECKING
import json

import geopandas as gpd
import pandas as pd
import shapely
import shapely.geometry as sg

from . import constants as cs
from . import helpers as hp

# Help mypy but avoid circular imports
if TYPE_CHECKING:
    from .feed import Feed


def append_dist_to_shapes(feed: "Feed") -> "Feed":
    """
    Calculate and append the optional ``shape_dist_traveled`` field in
    ``feed.shapes`` in terms of the distance units ``feed.dist_units``.
    Return the resulting Feed.

    As a benchmark, using this function on `this Portland feed
    <https://transitfeeds.com/p/trimet/43/1400947517>`_
    produces a ``shape_dist_traveled`` column that differs by at most
    0.016 km in absolute value from of the original values.
    """
    if feed.shapes is None:
        raise ValueError("This function requires the feed to have a shapes.txt file")

    feed = feed.copy()

    g = (
        feed.shapes.assign(
            geometry=lambda x: gpd.points_from_xy(x["shape_pt_lon"], x["shape_pt_lat"])
        )
        .pipe(gpd.GeoDataFrame, crs=cs.WGS84)
        .pipe(lambda x: x.to_crs(x.estimate_utm_crs()))
        .sort_values(["shape_id", "shape_pt_sequence"])
    )
    # Compute cumulative between successive points within shape
    g["prev_geom"] = g.groupby("shape_id")["geometry"].shift(1)
    g["dist_m"] = g.apply(
        lambda row: (
            row["geometry"].distance(row["prev_geom"])
            if pd.notnull(row["prev_geom"])
            else 0
        ),
        axis=1,
    )
    g["shape_dist_traveled"] = (
        g.groupby("shape_id")["dist_m"]
        .cumsum()
        .map(hp.get_convert_dist("m", feed.dist_units))
    )

    feed.shapes = g.drop(["geometry", "prev_geom", "dist_m"], axis=1)
    return feed


def geometrize_shapes(
    shapes: pd.DataFrame, *, use_utm: bool = False
) -> gpd.GeoDataFrame:
    """
    Given a GTFS shapes DataFrame, convert it to a GeoDataFrame of LineStrings
    and return the result, which will no longer have the columns
    ``'shape_pt_sequence'``, ``'shape_pt_lon'``,
    ``'shape_pt_lat'``, and ``'shape_dist_traveled'``.

    If ``use_utm``, then use local UTM coordinates for the geometries.
    """

    def my_agg(group):
        d = {}
        coords = group[["shape_pt_lon", "shape_pt_lat"]].values
        try:
            d["geometry"] = sg.LineString(coords)
        except shapely.errors.GEOSException:
            d["geometry"] = sg.Point(coords)

        return pd.Series(d)

    g = (
        shapes.sort_values(["shape_id", "shape_pt_sequence"])
        .groupby("shape_id", sort=False)
        .apply(my_agg, include_groups=False)
        .reset_index()
        .pipe(gpd.GeoDataFrame)
        .set_crs(cs.WGS84)
    )

    if use_utm:
        g = g.to_crs(g.estimate_utm_crs())

    return g


def ungeometrize_shapes(shapes_g: gpd.GeoDataFrame) -> pd.DataFrame:
    """
    The inverse of :func:`geometrize_shapes`.

    If ``shapes_g`` is in UTM coordinates (has a UTM CRS property),
    then convert those UTM coordinates back to WGS84 coordinates,
    which is the standard for a GTFS shapes table.
    """
    shapes_g = shapes_g.to_crs(cs.WGS84)

    F = []
    for index, row in shapes_g.iterrows():
        F.extend(
            [
                [row["shape_id"], i, x, y]
                for i, (x, y) in enumerate(row["geometry"].coords)
            ]
        )

    return pd.DataFrame(
        F,
        columns=[
            "shape_id",
            "shape_pt_sequence",
            "shape_pt_lon",
            "shape_pt_lat",
        ],
    ).astype({"shape_id": "string"})


def split_simple(
    shapes_g: gpd.GeoDataFrame, segmentize_m: float = 5
) -> gpd.GeoDataFrame:
    """
    Given GTFS shapes as a GeoDataFrame of the form output by :func:`geometrize_shapes`
    and possibly in a non-WGS84 CRS,
    split each non-simple LineString into maximal simple (non-self-intersecting)
    sub-LineStrings, and leave the simple LineStrings as is.
    Before splitting, segmentize (with Shapely's ``segmentize`` method)
    each non-simple LineString L by ``segmentize_m`` meters,
    which also sets the maximum gap size between L's simple sub-LineStrings.

    Return a GeoDataFrame in the CRS of ``shapes_g`` with the columns

    - ``'shape_id'``: a unique identifier of the original LineString L
    - ``'subshape_id'``: a unique identifier of a simple sub-LineString S of L
    - ``'subshape_sequence'``: integer; indicates the order of S when joining up
      all simple sub-LineStrings to form L
    - ``'subshape_length_m'``: the length of S in meters
    - ``'cum_length_m'``: the length S plus the lengths of sub-LineStrings of L
      that come before S; in meters
    - ``'geometry'``: LineString geometry corresponding to S

    Within each 'shape_id' group, the subshapes will be sorted increasingly by
    'subshape_sequence'.

    Note that by construction, for each given LineString L with k simple subLineStrings
    S_i, we have the inequalities

    length(L) - k * segmentize_m <= sum over i of length(S_i) <= length(L),

    where the lengths are expressed in meters.
    """

    def my_split(group):
        coords = group["geometry"].iat[0].coords
        segments = []
        n = len(coords)
        i = 0
        while i < n:
            # If only one coordinate remains, break  to
            # avoids making a degenerate LineString
            if i == n - 1:
                break
            # Start a binary search with at least two points
            lo, hi = i + 1, n - 1
            best = i + 1
            while lo <= hi:
                mid = (lo + hi) // 2
                candidate = sg.LineString(coords[i : mid + 1])
                if candidate.is_simple:
                    best = mid  # candidate is simple; try extending further.
                    lo = mid + 1
                else:
                    hi = mid - 1
            segments.append(sg.LineString(coords[i : best + 1]))
            if best == n - 1:
                break
            # Start next segment after the current best segment
            i = best + 1
        return pd.DataFrame({"geometry": segments})

    crs = shapes_g.crs
    utm_crs = shapes_g.estimate_utm_crs()
    g = shapes_g.assign(is_simple=lambda x: x.is_simple).to_crs(utm_crs)

    # Simple shapes don't need splitting
    g0 = g.loc[lambda x: x["is_simple"]].assign(
        subshape_id=lambda x: x["shape_id"].astype(str) + "-0",
        subshape_sequence=0,
        subshape_length_m=lambda x: x.length,
        cum_length_m=lambda x: x["subshape_length_m"],
    )
    # Split the non-simple shapes
    g1 = (
        g.loc[lambda x: ~x["is_simple"]]
        .assign(geometry=lambda x: x.segmentize(segmentize_m))
        .groupby("shape_id", sort=False)
        .apply(my_split, include_groups=False)
        .reset_index()
        .rename(columns={"level_1": "subshape_sequence"})
        .assign(
            subshape_id=lambda x: x["shape_id"].str.cat(
                x["subshape_sequence"].astype(str), sep="-"
            )
        )
        .pipe(gpd.GeoDataFrame)
        .set_crs(utm_crs)
        .assign(
            subshape_length_m=lambda x: x.length,
            cum_length_m=lambda x: x.groupby("shape_id")["subshape_length_m"].cumsum(),
        )
    )
    return (
        pd.concat([g0, g1])
        .to_crs(crs)
        .filter(
            [
                "subshape_id",
                "subshape_sequence",
                "shape_id",
                "subshape_length_m",
                "cum_length_m",
                "geometry",
            ]
        )
        .sort_values(["subshape_id", "subshape_sequence"], ignore_index=True)
    )


def get_shapes(
    feed: "Feed", *, as_gdf: bool = False, use_utm: bool = False
) -> gpd.DataFrame | None:
    """
    Get the shapes DataFrame for the given feed, which could be ``None``.
    If ``as_gdf``, then return it as GeoDataFrame with a 'geometry' column
    of linestrings and no 'shape_pt_sequence', 'shape_pt_lon', 'shape_pt_lat',
    'shape_dist_traveled' columns.
    The GeoDataFrame will have a UTM CRS if ``use_utm``; otherwise it will have a
    WGS84 CRS.
    """
    f = feed.shapes
    if f is not None and as_gdf:
        f = geometrize_shapes(f, use_utm=use_utm)
    return f


def build_geometry_by_shape(
    feed: "Feed", shape_ids: Iterable[str] | None = None, *, use_utm: bool = False
) -> dict:
    """
    Return a dictionary of the form
    <shape ID> -> <Shapely LineString representing shape>.
    If the Feed has no shapes, then return the empty dictionary.
    If ``use_utm``, then use local UTM coordinates; otherwise, use WGS84 coordinates.
    """
    if feed.shapes is None:
        return dict()

    g = get_shapes(feed, as_gdf=True, use_utm=use_utm)
    if shape_ids is not None:
        g = g.loc[lambda x: x["shape_id"].isin(shape_ids)]
    return dict(g[["shape_id", "geometry"]].values)


def shapes_to_geojson(feed: "Feed", shape_ids: Iterable[str] | None = None) -> dict:
    """
    Return a GeoJSON FeatureCollection of LineString features
    representing ``feed.shapes``.
    If the Feed has no shapes, then the features will be an empty list.
    The coordinates reference system is the default one for GeoJSON,
    namely WGS84.

    If an iterable of shape IDs is given, then subset to those shapes.
    If the subset is empty, then return a FeatureCollection with an empty list of
    features.
    """
    g = get_shapes(feed, as_gdf=True)
    if shape_ids is not None:
        g = g.loc[lambda x: x["shape_id"].isin(shape_ids)]
    if g is None or g.empty:
        result = {
            "type": "FeatureCollection",
            "features": [],
        }
    else:
        result = hp.drop_feature_ids(json.loads(g.to_json()))
    return result


def get_shapes_intersecting_geometry(
    feed: "Feed",
    geometry: sg.base.BaseGeometry,
    shapes_g: gpd.GeoDataFrame | None = None,
    *,
    as_gdf: bool = False,
) -> pd.DataFrame | None:
    """
    If the Feed has no shapes, then return None.
    Otherwise, return the subset of ``feed.shapes`` that contains all shapes that
    intersect the given Shapely WGS84 geometry, e.g. a Polygon or LineString.

    If ``as_gdf``, then return the shapes as a GeoDataFrame.
    Specifying ``shapes_g`` will skip the first step of the
    algorithm, namely, geometrizing ``feed.shapes``.
    """
    if feed.shapes is None:
        return None

    if shapes_g is not None:
        g = shapes_g.copy()
    else:
        g = get_shapes(feed, as_gdf=True)

    cols = g.columns
    g["hit"] = g["geometry"].intersects(geometry)
    g = g.loc[lambda x: x["hit"]].filter(cols)

    if as_gdf:
        result = g
    else:
        result = ungeometrize_shapes(g)

    return result



import pytest
from pathlib import Path
import shutil
import tempfile

import pandas as pd
from pandas.testing import assert_frame_equal
import numpy as np

from .context import gtfs_kit, DATA_DIR
from gtfs_kit import feed as gkf
from gtfs_kit import constants as cs


def test_feed():
    feed = gkf.Feed(agency=pd.DataFrame(), dist_units="km")
    for key in cs.FEED_ATTRS:
        val = getattr(feed, key)
        if key == "dist_units":
            assert val == "km"
        elif key == "agency":
            assert isinstance(val, pd.DataFrame)
        else:
            assert val is None


def test_str():
    feed = gkf.Feed(agency=pd.DataFrame(), dist_units="km")
    assert isinstance(str(feed), str)


def test_eq():
    assert gkf.Feed(dist_units="m") == gkf.Feed(dist_units="m")

    feed1 = gkf.Feed(
        dist_units="m",
        stops=pd.DataFrame([[1, 2], [3, 4]], columns=["a", "b"]),
    )
    assert feed1 == feed1

    feed2 = gkf.Feed(
        dist_units="m",
        stops=pd.DataFrame([[4, 3], [2, 1]], columns=["b", "a"]),
    )
    assert feed1 == feed2

    feed2 = gkf.Feed(
        dist_units="m",
        stops=pd.DataFrame([[3, 4], [2, 1]], columns=["b", "a"]),
    )
    assert feed1 != feed2

    feed2 = gkf.Feed(
        dist_units="m",
        stops=pd.DataFrame([[4, 3], [2, 1]], columns=["b", "a"]),
    )
    assert feed1 == feed2

    feed2 = gkf.Feed(dist_units="mi", stops=feed1.stops)
    assert feed1 != feed2


def test_copy():
    feed1 = gkf.read_feed(DATA_DIR / "sample_gtfs.zip", dist_units="km")
    feed2 = feed1.copy()

    # Check attributes
    for key in cs.FEED_ATTRS:
        val = getattr(feed2, key)
        expect_val = getattr(feed1, key)
        if isinstance(val, pd.DataFrame):
            assert_frame_equal(val, expect_val)
        elif isinstance(val, pd.core.groupby.DataFrameGroupBy):
            assert val.groups == expect_val.groups
        else:
            assert val == expect_val


# --------------------------------------------
# Test functions about inputs and outputs
# --------------------------------------------
def test_list_feed():
    # Bad path
    with pytest.raises(ValueError):
        gkf.list_feed("bad_path!")

    for path in [DATA_DIR / "sample_gtfs.zip", DATA_DIR / "sample_gtfs"]:
        f = gkf.list_feed(path)
        assert isinstance(f, pd.DataFrame)
        assert set(f.columns) == {"file_name", "file_size"}
        assert f.shape[0] in [12, 13]


def test_read_feed():
    # Bad path
    with pytest.raises(ValueError):
        gkf.read_feed("bad_path!", dist_units="km")

    # Bad dist_units
    with pytest.raises(ValueError):
        gkf.read_feed(DATA_DIR / "sample_gtfs.zip", dist_units="bingo")

    # Requires dist_units
    with pytest.raises(TypeError):
        gkf.read_feed(path=DATA_DIR / "sample_gtfs.zip")

    # Success
    feed = gkf.read_feed(DATA_DIR / "sample_gtfs.zip", dist_units="m")

    # Success
    feed = gkf.read_feed(DATA_DIR / "sample_gtfs", dist_units="m")

    # Feed should have None feed_info table
    assert feed.feed_info is None


def test_to_file():
    feed1 = gkf.read_feed(DATA_DIR / "sample_gtfs.zip", dist_units="km")

    # Export feed1, import it as feed2, and then test equality
    for out_path in [DATA_DIR / "bingo.zip", DATA_DIR / "bingo"]:
        feed1.to_file(out_path)
        feed2 = gkf.read_feed(out_path, "km")
        assert feed1 == feed2
        try:
            out_path.unlink()
        except Exception:
            shutil.rmtree(str(out_path))

    # Test that integer columns with NaNs get output properly.
    feed3 = gkf.read_feed(DATA_DIR / "sample_gtfs.zip", dist_units="km")
    f = feed3.trips.copy()
    f.loc[0, "direction_id"] = np.nan
    f.loc[1, "direction_id"] = 1
    f.loc[2, "direction_id"] = 0
    feed3.trips = f
    q = DATA_DIR / "bingo.zip"
    feed3.to_file(q)

    tmp_dir = tempfile.TemporaryDirectory()
    shutil.unpack_archive(str(q), tmp_dir.name, "zip")
    qq = Path(tmp_dir.name) / "trips.txt"
    t = pd.read_csv(qq, dtype={"direction_id": "Int8"})
    assert t[~t["direction_id"].isin([pd.NA, 0, 1])].empty
    tmp_dir.cleanup()
    q.unlink()
